# type: ignore
import os
import tempfile
import yaml

from e3.env import Env
from e3.fs import rm
from e3.os.fs import cd, which
from e3.config import Config

import pytest

# When the variable RESULTS_DIR is set to
# an existing directory, the testsuite will
# generate results file in "anod" format
RESULTS_DIR = os.environ.get("RESULTS_DIR")


class RequirementCoverage:
    """Track requirements <-> tests."""

    output_filename = None
    results = {}

    @classmethod
    def dump(cls):
        if cls.output_filename:
            with open(cls.output_filename, "w") as f:
                yaml.safe_dump(cls.results, f)


@pytest.fixture(autouse=True)
def env_protect(request):
    """Protection against environment change.

    The fixture is enabled for all tests and does the following:

    * store/restore env between each tests
    * create a temporary directory and do a cd to it before each
      test. The directory is automatically removed when test ends
    """
    Env().store()
    tempd = tempfile.mkdtemp()
    cd(tempd)
    Config.data = {}

    def restore_env():
        Env().restore()
        rm(tempd, True)

    request.addfinalizer(restore_env)


def pytest_configure(config):
    try:
        RequirementCoverage.output_filename = config.getoption(
            "requirement_coverage_report"
        )
    except ValueError:
        # Option not defined.
        pass


def require_vcs(prog, request):
    """Require svn or git to run the test.

    When in "CI" mode, a missing svn or git generates an error. In other
    modes the test is just skipped.
    :param prog: either "svn" or "git"
    """
    if not which(prog):
        if request.config.getoption("ci"):
            pytest.fail(f"{prog} not available")
        else:
            pytest.skip(f"{prog} not available")


@pytest.fixture(autouse=True)
@pytest.mark.usefixtures("git")
def require_git(request):
    """Require git."""
    marker = request.node.get_closest_marker("git")
    if marker:
        return require_vcs("git", request)


@pytest.fixture
def git(request):
    """Require git."""
    return require_vcs("git", request)


@pytest.fixture(autouse=True)
@pytest.mark.usefixtures("svn")
def require_svn(request):
    """Require svn."""
    marker = request.node.get_closest_marker("svn")
    if marker:
        return require_vcs("svn", request)


def pytest_itemcollected(item):
    """Keep track of all test linked to a requirement."""
    doc = item.obj.__doc__
    if RequirementCoverage.output_filename and doc:
        for line in item.obj.__doc__.splitlines():
            line = line.strip().strip(".")
            if line.startswith("REQ-"):
                RequirementCoverage.results[item.obj.__name__] = line


def pytest_collectreport(report):
    """Output requirement coverage report."""
    RequirementCoverage.dump()


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """Generate results file that can be used by anod."""
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    if not RESULTS_DIR or not os.path.isdir(RESULTS_DIR):
        return

    # we only look at actual test calls, not setup/teardown
    if rep.when == "call":
        outcome = rep.outcome.upper()
        test_name = rep.nodeid.replace("/", ".").replace("::", "--")
        if rep.longreprtext:
            with open(os.path.join(RESULTS_DIR, f"{test_name}.diff"), "w") as f:
                f.write(rep.longreprtext)

        with open(os.path.join(RESULTS_DIR, "results"), "a") as f:
            f.write(f"{test_name}:{outcome}\n")



"""Decorators.

to enable/disable a function, to memoize a function results...
"""


from __future__ import annotations

from functools import partial

from typing import TYPE_CHECKING
from warnings import warn

if TYPE_CHECKING:
    from typing import Any, Callable, Dict


def enabled(func: Callable) -> Callable:
    """no-op. Do not change function behaviour.

    If you write the following code::

        @enabled
        def foo():
            print("I'm foo")

    Then calling ``foo()`` will return "I'm foo"

    :param func: function to decorate
    """
    return func


def disabled(func: Callable) -> Callable:
    """Disable the provided function, and does nothing.

    If you write the following code::

        @disabled
        def foo():
            print("I'm foo")

    Then calling ``foo()`` will return None

    :param func: function to decorate
    """
    del func

    def empty_func(*args, **kargs):
        del args, kargs

    return empty_func


class memoize:
    """Memoize function return values.

    Avoid repeating the calculation of results for previously-processed
    inputs.

    If you write the following code::

        import random
        @memoize
        def long_computation(r):
            del r
            return random.random()

    Then you will have::

       long_computation(42) == long_computation(42)

    Calling the same function twice with the same paramaters returns the same
    result.

    Calling the function with the special keyword argument reset_cache=True
    force a call to the decorated function, skipping the cache.

    No keyword argument can be passed to the decorated function.
    """

    def __init__(self, func: Callable):
        """Initialize the decorator.

        :param func: function to decorate
        """
        self.func = func
        self.cache: Dict[tuple, Any] = {}
        warn(
            "this decorator will be removed in a later version of e3-core, "
            "please use functools.lru_cache instead",
            category=DeprecationWarning,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Return the cache value if exist, else call func."""
        if kwargs:
            if len(kwargs) == 1 and kwargs.get("reset_cache"):
                # special handling of reset_cache, clean the cache
                if args in self.cache:
                    del self.cache[args]
            else:
                raise TypeError("memoize does not support keyword arguments")
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # non-hashable arguments, skip the cache
            return self.func(*args)

    def __repr__(self) -> str:
        """Return the function's docstring."""
        return self.func.__doc__ or ""

    def __get__(self, obj, objtype):
        """Support instance methods."""
        del objtype
        return partial(self.__call__, obj)



import os
import re
import sys

import e3.env
import e3.os.process

import pytest


def test_main():
    assert (
        e3.env.Env().build.platform
        in e3.os.process.Run(["e3", "--platform-info=build"]).out
    )


def test_main_config():
    os.environ["E3_CONFIG"] = "e3.toml"
    assert "pretty: True" in e3.os.process.Run(["e3", "--show-config"]).out

    with open("e3.toml", "w") as f:
        f.write("[log]\npretty = false\n")
    assert "pretty: False" in e3.os.process.Run(["e3", "--show-config"]).out

    # Verify that invalid config field is ignored
    with open("e3.toml", "w") as f:
        f.write('[log]\npretty = "false"\nstream_fmt = "%(message)s"')
    out = e3.os.process.Run(["e3", "--show-config"]).out
    assert "pretty: True" in out
    assert "stream_fmt: '%(message)s'" in out
    assert "type of log.pretty must be bool" in out

    # And finally check that invalid toml are discarded
    with open("e3.toml", "w") as f:
        f.write("this is an invalid toml content")
    assert "pretty: True" in e3.os.process.Run(["e3", "--show-config"]).out


def test_mainprog():
    with open("mymain.py", "w") as f:
        f.write(
            "\n".join(
                (
                    "#!/usr/bin/env python",
                    "from e3.main import Main",
                    "import os",
                    'm = Main(name="testmain")',
                    "m.parse_args()",
                    "m.argument_parser.print_usage()",
                )
            )
        )
    p = e3.os.process.Run([sys.executable, "mymain.py", "--nocolor"])
    assert "mymain" in p.out


def test_mainprog_with_console_logs():
    with open("mymain.py", "w") as f:
        f.write(
            "\n".join(
                (
                    "#!/usr/bin/env python",
                    "from e3.main import Main",
                    "import os",
                    'm = Main(name="testmain")',
                    "m.parse_args()",
                    "import logging",
                    'logging.debug("this is an info line")',
                    'logging.debug("this is a debug line")',
                )
            )
        )
    p = e3.os.process.Run(
        [sys.executable, "mymain.py", "-v", "--console-logs=mymain", "--nocolor"]
    )

    assert re.search(
        r"^mymain:.*:.*: DEBUG    this is an info line\r?\n"
        "mymain:.*:.* DEBUG    this is a debug line",
        p.out,
    )


@pytest.mark.skipif(
    sys.platform in ("win32", "sunos5"),
    reason="Signal handler not set on windows." " Bug in signal handling in solaris",
)
def test_sigterm():
    with open("mymain.py", "w") as f:
        f.write(
            "\n".join(
                (
                    "#!/usr/bin/env python",
                    "from e3.main import Main",
                    "import os",
                    "import signal",
                    "m = Main()",
                    "m.parse_args()",
                    "import time",
                    "time.sleep(10)",
                )
            )
        )
    p = e3.os.process.Run([sys.executable, "mymain.py", "--nocolor"], timeout=1)
    assert "SIGTERM received" in p.out



"""Utility functions related to json."""


from __future__ import annotations

import json
import os

import e3.error

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


class JsonError(e3.error.E3Error):
    pass


def dump_to_json_file(path: str, obj: Any) -> None:
    """Dump a Python object to a json file.

    :param path: path to the json file
    :param obj: a Python object that can serialized to JSON
    """
    with open(path, "w") as fd:
        json.dump(obj, fd, indent=2)


def load_from_json_file(
    path: str, default: Any = None, ignore_non_existing: bool = True
) -> Any:
    """Load a Python object from a JSON file.

    :param path: json file path
    :param default: default value returned if ignore_non_existing is True and
        the specified file does not exist.
    :param ignore_non_existing: if False raise JsonError if the file does not
        exist, otherwise return default value
    :return: a Python object
    """
    if os.path.isfile(path):
        with open(path) as fd:
            content = json.load(fd)
        return content
    else:
        if ignore_non_existing:
            return default
        else:
            raise JsonError("json file %s does not exist" % path)



import os

import e3.anod.driver
import e3.anod.sandbox
import e3.anod.spec

import pytest


def test_simple_driver():
    sandbox = e3.anod.sandbox.SandBox()

    class Simple(e3.anod.spec.Anod):
        @e3.anod.spec.Anod.primitive()
        def download(self):
            pass

    with pytest.raises(e3.anod.spec.AnodError):
        anod_instance = Simple(qualifier="", kind="build")
        anod_instance.sandbox = None
        e3.anod.driver.AnodDriver(anod_instance=anod_instance, store=None).activate(
            sandbox, None
        )

    sandbox.root_dir = os.getcwd()
    anod_instance = Simple(qualifier="", kind="build")
    anod_instance.sandbox = sandbox
    driver = e3.anod.driver.AnodDriver(anod_instance=anod_instance, store=None)

    assert driver.call("why") is False
    with pytest.raises(e3.anod.spec.AnodError):
        driver.download()


def test_deps_driver():
    class Deps(e3.anod.spec.Anod):
        build_deps = [e3.anod.spec.Anod.Dependency(name="parent")]

        @e3.anod.spec.Anod.primitive()
        def build(self):
            return self.deps["parent"].parent_info

    sandbox = e3.anod.sandbox.SandBox()
    sandbox.root_dir = os.getcwd()
    anod_instance = Deps(qualifier="", kind="build")
    anod_instance.sandbox = sandbox

    spec_dir = os.path.join(os.path.dirname(__file__), "data")
    spec_repo = e3.anod.loader.AnodSpecRepository(spec_dir)

    e3.anod.driver.AnodDriver(anod_instance=anod_instance, store=None).activate(
        sandbox, spec_repo
    )

    anod_instance.build_space.create()
    assert anod_instance.build() == "from_parent"



from __future__ import annotations

import ctypes
import sys
import time
from ctypes import POINTER, Structure, c_wchar_p
from ctypes import cast as ctypes_cast
from ctypes import create_unicode_buffer, pointer, sizeof
from ctypes.wintypes import (
    BOOL,
    BOOLEAN,
    DWORD,
    HANDLE,
    INT,
    LARGE_INTEGER,
    LONG,
    LPVOID,
    LPWSTR,
    ULONG,
    USHORT,
)
from datetime import datetime
from typing import List, Optional

from e3.error import E3Error

NTSTATUS = LONG


class FileAttribute(Structure):
    """List of file attributes constants."""

    READONLY = 0x01
    HIDDEN = 0x02
    SYSTEM = 0x04
    DIRECTORY = 0x10
    ARCHIVE = 0x20
    DEVICE = 0x40
    NORMAL = 0x80
    TEMPORARY = 0x100
    SPARSE_FILE = 0x200
    REPARSE_POINT = 0x400
    COMPRESSED = 0x800
    OFFLINE = 0x1000
    NOT_CONTENT_INDEXED = 0x2000
    ENCRYPTED = 0x4000
    INTEGRITY_STREAM = 0x8000
    VIRTUAL = 0x10000
    NO_SCRUB_DATA = 0x20000

    _fields_ = [("attr", ULONG)]

    def __str__(self):
        result = []
        for k in FileAttribute.__dict__:
            if not k.startswith("_") and k.isupper():
                if FileAttribute.__dict__[k] & self.attr > 0:
                    result.append(k)
        return ",".join(result)


class Access:
    """Desired Access constants."""

    LIST_DIRECTORY = 0x0001
    READ_DATA = 0x0001
    ADD_FILE = 0x0002
    WRITE_DATA = 0x0002
    ADD_SUBDIR = 0x0004
    APPEND_DATA = 0x0004
    READ_EA = 0x0008
    WRITE_EA = 0x0010
    EXECUTE = 0x0020
    TRAVERSE = 0x0020
    DELETE_CHILD = 0x0040
    READ_ATTRS = 0x0080
    WRITE_ATTRS = 0x0100
    DELETE = 0x010000
    SYNCHRONIZE = 0x100000
    PROCESS_QUERY_INFORMATION = 0x0400


class Share:
    """Share Access constants."""

    NOTHING = 0x00
    READ = 0x01
    WRITE = 0x02
    DELETE = 0x04
    ALL = 0x07


class OpenOptions:
    """File Open options."""

    BACKUP_INTENT = 0x00004000
    SYNCHRONOUS_IO_NON_ALERT = 0x00000020
    DELETE_ON_CLOSE = 0x00001000


class Wait:
    """Constants for WaitFor* functions."""

    OBJECT = 0x0000
    ABANDONED = 0x080
    TIMEOUT = 0x0102
    FAILED = 0xFFFFFFFF
    INFINITE = 0xFFFFFFFF


class Status:
    """Error constants."""

    ACCESS_DENIED = 0xC0000022
    OBJECT_NAME_NOT_FOUND = 0xC0000034
    OBJECT_PATH_NOT_FOUND = 0xC000003A
    SHARING_VIOLATION = 0xC0000043
    DELETE_PENDING = 0xC0000056
    DIRECTORY_NOT_EMPTY = 0xC0000101
    CANNOT_DELETE = 0xC0000121
    NO_MORE_FILES = 0x80000006

    msgs = {
        0xC0000022: "access denied",
        0xC0000034: "object name not found",
        0xC0000043: "sharing violation",
        0xC0000056: "delete pending",
        0xC000003A: "object path not found",
        0xC0000101: "directory is not empty",
        0xC0000121: "cannot delete",
    }


class IOStatusBlock(Structure):
    """Map IO_STATUS_BLOCK structure."""

    _fields_ = [("status", NTSTATUS), ("information", POINTER(ULONG))]


class UnicodeString(Structure):
    """Map UNICODE_STRING structure."""

    _fields_ = [("length", USHORT), ("maximum_length", USHORT), ("buffer", LPWSTR)]

    def __init__(self, value: Optional[str] = None, max_length: int = 0):
        strbuf = None
        length = 0
        if value is not None or max_length > 0:
            length = len(value) if value is not None else 0
            max_length = max(length, max_length)
            if value:
                strbuf = create_unicode_buffer(value, max_length)
            else:
                strbuf = create_unicode_buffer(max_length)

        ctypes_strbuf = ctypes_cast(strbuf, LPWSTR)  # type: ignore

        Structure.__init__(self, length * 2, max_length * 2, ctypes_strbuf)

    def __len__(self) -> int:
        return self.length


class FileTime(Structure):
    """Map FILETIME structure."""

    _fields_ = [("filetime", LARGE_INTEGER)]

    def __init__(self, t):
        timestamp = (t - datetime(1970, 1, 1)).total_seconds()
        timestamp = (int(timestamp) + 11644473600) * 10000000
        Structure.__init__(self, timestamp)

    @property
    def as_datetime(self):
        try:
            return datetime.fromtimestamp(self.filetime // 10000000 - 11644473600)
        except ValueError as err:  # defensive code
            # Add some information to ease debugging
            raise ValueError(f"filetime '{self.filetime}' failed with {err}")

    def __str__(self):
        try:
            return str(time.ctime(self.filetime // 10000000 - 11644473600))
        except ValueError:  # defensive code
            return "none"


class FileInfo:
    """Declaration of structures returned by QueryInformationFile."""

    class Names:
        class_id = 12

    class Disposition(Structure):
        _fields_ = [("delete_file", BOOLEAN)]
        class_id = 13

    class Internal(Structure):
        _fields_ = [("index_number", LARGE_INTEGER)]
        class_id = 6

    class Rename(Structure):
        _fields_ = []  # type: List
        class_id = 10

    class Basic(Structure):
        _fields_ = [
            ("creation_time", FileTime),
            ("last_access_time", FileTime),
            ("last_write_time", FileTime),
            ("change_time", FileTime),
            ("file_attributes", FileAttribute),
        ]
        class_id = 4

        def __init__(self) -> None:
            Structure.__init__(self)
            self.file_attributes = FileAttribute(0)

        def __str__(self) -> str:
            result = "creation_time:    %s\n" % self.creation_time
            result += "last_access_time: %s\n" % self.last_access_time
            result += "last_write_time:  %s\n" % self.last_write_time
            result += "change_time:      %s\n" % self.change_time
            result += "file_attributes:  %s" % self.file_attributes
            return result


class ProcessInfo:
    """Declaration of structure returned by QueryInformationProcess."""

    class Basic(Structure):
        class_id = 0

        _fields_ = [
            ("exit_status", NTSTATUS),
            ("peb", LPVOID),
            ("affinity_mask", LPVOID),
            ("base_priority", LONG),
            ("pid", LPVOID),
            ("ppid", LPVOID),
        ]

    STILL_ACTIVE = 259


class ObjectAttributes(Structure):
    """OBJECT_ATTRIBUTES structure."""

    OBJ_CASE_INSENSITIVE = 0x00000040

    _fields_ = [
        ("length", ULONG),
        ("root_directory", HANDLE),
        ("object_name", POINTER(UnicodeString)),
        ("attributes", ULONG),
        ("security_descriptor", LPVOID),
        ("security_quality_of_service", LPVOID),
    ]

    def __init__(self, name: UnicodeString, parent: Optional[HANDLE] = None):
        """Initialize ObjectAttributes.

        :param name: full path to the file if parent is None else filename
            inside parent directory
        :param parent: handle of the parent directory
        """
        Structure.__init__(
            self,
            sizeof(ObjectAttributes),
            parent,
            pointer(name),
            self.OBJ_CASE_INSENSITIVE,
            None,
            None,
        )


# Declare the Win32 functions return types and signature
class NT:
    Sleep = None
    GetVolumePathName = None
    SetInformationFile = None
    QueryInformationFile = None
    QueryAttributesFile = None
    OpenFile = None
    QueryDirectoryFile = None
    Close = None
    QueryInformationProcess = None
    WaitForMultipleObjects = None
    OpenProcess = None

    @classmethod
    def init_api(cls):
        kernel32 = ctypes.windll.kernel32
        ntdll = ctypes.windll.ntdll

        cls.GetVolumePathName = kernel32.GetVolumePathNameW
        cls.GetVolumePathName.restype = BOOL
        cls.GetVolumePathName.argtypes = [c_wchar_p, c_wchar_p, DWORD]
        cls.Sleep = kernel32.Sleep
        cls.Sleep.argtypes = [DWORD]

        cls.SetInformationFile = ntdll.NtSetInformationFile
        cls.SetInformationFile.restype = NTSTATUS
        cls.SetInformationFile.argtypes = [
            HANDLE,
            POINTER(IOStatusBlock),
            LPVOID,
            ULONG,
            INT,
        ]
        cls.QueryInformationFile = ntdll.NtQueryInformationFile
        cls.QueryInformationFile.restype = NTSTATUS
        cls.QueryInformationFile.argtypes = [
            HANDLE,
            POINTER(IOStatusBlock),
            LPVOID,
            ULONG,
            INT,
        ]
        cls.QueryAttributesFile = ntdll.NtQueryAttributesFile
        cls.QueryAttributesFile.restype = NTSTATUS
        cls.QueryAttributesFile.argtypes = [
            POINTER(ObjectAttributes),
            POINTER(FileInfo.Basic),
        ]
        cls.OpenFile = ntdll.NtOpenFile
        cls.OpenFile.restype = NTSTATUS
        cls.OpenFile.argtypes = [
            POINTER(HANDLE),
            DWORD,
            POINTER(ObjectAttributes),
            POINTER(IOStatusBlock),
            ULONG,
            ULONG,
        ]
        cls.QueryDirectoryFile = ntdll.NtQueryDirectoryFile
        cls.QueryDirectoryFile.restype = NTSTATUS
        cls.QueryDirectoryFile.argtypes = [
            HANDLE,
            HANDLE,
            LPVOID,
            LPVOID,
            POINTER(IOStatusBlock),
            LPVOID,
            ULONG,
            INT,
            BOOLEAN,
            POINTER(UnicodeString),
            BOOLEAN,
        ]
        cls.Close = ntdll.NtClose
        cls.Close.argtypes = [HANDLE]
        cls.QueryInformationProcess = ntdll.NtQueryInformationProcess
        cls.QueryInformationProcess.restype = NTSTATUS
        cls.QueryInformationProcess.argtypes = [
            HANDLE,
            INT,
            POINTER(ProcessInfo.Basic),
            ULONG,
            LPVOID,
        ]
        cls.WaitForMultipleObjects = kernel32.WaitForMultipleObjects
        cls.WaitForMultipleObjects.restype = DWORD
        cls.WaitForMultipleObjects.argtypes = [DWORD, POINTER(HANDLE), BOOLEAN, DWORD]
        cls.OpenProcess = kernel32.OpenProcess
        cls.OpenProcess.restype = HANDLE
        cls.OpenProcess.argtypes = [DWORD, BOOL, DWORD]


if sys.platform == "win32":
    NT.init_api()


class NTException(E3Error):
    def __init__(self, status: int, message: str, origin: Optional[str] = None):
        self.status = status
        if self.status < 0:
            self.status += 2 ** 32
        E3Error.__init__(self, message, origin=origin)

    def __str__(self):
        return E3Error.__str__(self) + "(status={:X} '{}')".format(
            self.status, Status.msgs.get(self.status, "unknown"),
        )
